# Copyright 2021 Florian Fischer
"""Server subprocess related code"""

import logging
from pathlib import Path
import subprocess

from .globalvars import KILL_CMD, TERMINATION_TIME
from .util import cmd_run, prepare_env

log = logging.getLogger(__name__)

SHUTDOWN_CMD = 'nc {host} 12345'

SERVER_OUT = '{run}.out'
SERVER_ERR = '{run}.err'


class Server(subprocess.Popen):
    """Wrapper around a server subprocess"""
    def __init__(self,
                 name: str,
                 cmd: str,
                 run: int,
                 bench_dir: Path,
                 env=None,
                 measure_cmd=None,
                 remote_cmd=None,
                 host=None):
        self.name = name
        self.cmd = cmd
        self.bin = cmd.split()[0]
        self.remote_cmd = remote_cmd

        if measure_cmd:
            self.cmd = f'{measure_cmd} {cmd}'

        if remote_cmd:
            self.cmd = f'{remote_cmd} {cmd}'
            assert host

        self.host = host

        out_file = bench_dir / SERVER_OUT.format(run=run)
        err_file = bench_dir / SERVER_ERR.format(run=run)

        log.debug(
            'Creating %s server: %s', name, {
                'cmd': self.cmd,
                'out': out_file,
                'err': err_file,
                'env': env,
                'host': host or 'localhost'
            })

        self.fout = open(out_file, 'w')
        self.ferr = open(err_file, 'w')

        server_env = prepare_env(env or {})

        super().__init__(cmd.split(),
                         stdout=self.fout,
                         stderr=self.ferr,
                         env=server_env,
                         text=True)  # type: ignore

    def shutdown(self) -> bool:
        """Gracfully shutdown a running server"""
        if self.poll() is not None:
            return True

        shutdown_cmd = SHUTDOWN_CMD.format(host=self.host or 'localhost')
        log.debug('Shutting down %s server by running: %s', self.name,
                  shutdown_cmd)
        # try to shutdown by sending "quit" to the server
        try:
            subprocess.run(shutdown_cmd.split(),
                           input=b'quit\n',
                           check=True,
                           timeout=TERMINATION_TIME)

            # wait for the server to terminate
            self.wait(TERMINATION_TIME)
            return True

        except subprocess.CalledProcessError as run_err:
            log.warning('%s terminated unsuccessful: %s', shutdown_cmd,
                        run_err)

        except subprocess.TimeoutExpired as timeout_err:
            log.warning('%s termination timeout: %s', shutdown_cmd,
                        timeout_err)

        return False

    def kill(self):
        """Kill a running server"""
        if self.poll() is not None:
            return

        # kill the server using pkill(1) -9
        kill_cmd = KILL_CMD.format(proc=self.bin)
        if self.remote_cmd:
            kill_cmd = f'{self.remote_cmd} {kill_cmd}'

        cmd_run(kill_cmd, log, check=False)
        self.wait(TERMINATION_TIME)

    def __del__(self):
        # the server is still alive
        if self.poll() is None:
            self.kill()

        if not self.shutdown():
            self.kill()

        self.fout.close()
        self.ferr.close()
