#!/usr/bin/env python3

import argparse
import csv
import os
from pathlib import Path
import subprocess
import sys

SAR_CMD = 'sar -f {sar_file} -n DEV --iface=eth0 | tail -1'

def calc_avgs(datadir):
    datadir = Path(datadir)
    implementaions = [d for d in datadir.iterdir() if d.is_dir()]

    # dict holding the avgs per implementation, processes, connection and size
    data = {}

    for implementation in implementaions:
        impl_dir = datadir / implementation
        data_files = impl_dir.glob(f'*.csv')
        for data_file in data_files:
            procs, cons, size, run = data_file.stem.split('.')

            procs, cons, size, run = int(procs), int(cons), int(size), int(run)
            key_tuple = (implementation.name, procs, cons, size)

            with open(data_file, 'r') as csvfile:
                r = csv.DictReader(csvfile, delimiter=';', skipinitialspace=True, lineterminator='\n')
                avg_d = data.setdefault(key_tuple, {})

                #get io data
                sar_file = Path(data_file.stem + '.sar.data')
                if sar_file.exists():
                    sar_cmd = SAR_CMD.format(sar_file)
                    io_avgs = subprocess.check_output(sar_cmd, shell=True).strip().decode()
                    io_avgs = io_avgs.split()

                    avg_io_key = 'ifutil'
                    if not avg_io_key in avg_d:
                        avg_d[avg_io_key] = io_avgs[-1]
                    else:
                        avg_d[avg_io_key] += io_avgs[-1]
                        avg_d[avg_io_key] /= 2

                avg_ns = 0
                avg_lat = 0
                avg_iops = 0
                # process averages
                for n, d in enumerate(r):
                    it, cons, proc_ns = int(d['iterations']), int(d['connections']), int(d['time[ns]'])
                    echos = cons * it

                    if avg_ns == 0:
                        avg_ns = proc_ns
                    else:
                        avg_ns += proc_ns
                        avg_ns /= 2

                    l_avg_lat = proc_ns / echos
                    if avg_lat == 0:
                        avg_lat = l_avg_lat
                    else:
                        avg_lat += l_avg_lat
                        avg_lat /= 2

                    # 1 second / avg time an echo took
                    #iops = (1* 10**9) / l_avg_lat
                    iops = echos / (proc_ns * 10**-9)
                    if avg_iops == 0:
                        avg_iops = iops
                    else:
                        avg_iops += iops
                        avg_iops /= 2

                if n + 1 != procs:
                    print(f'{key_tuple}: not all processes {n}/{procs} reported a result')
                    data[key_tuple] = {'time': 'nan', 'latency': 'nan', 'iops': 'nan', 'ifutil': 'nan'}
                    continue


                # inter run averages
                avg_ns_key = 'time'
                if not avg_ns_key in d:
                    avg_d[avg_ns_key] = avg_ns
                else:
                    avg_d[avg_ns_key] += avg_ns
                    avg_d[avg_ns_key] /= 2

                avg_lat_key = 'latency'
                if not avg_lat_key in d:
                    avg_d[avg_lat_key] = avg_lat
                else:
                    avg_d[avg_lat_key] += avg_lat
                    avg_d[avg_lat_key] /= 2

                iops_key = 'iops'
                total_iops = avg_iops * procs
                if not iops_key in d:
                    avg_d[iops_key] = total_iops
                else:
                    avg_d[iops_key] += total_iops
                    avg_d[iops_key] /= 2


    with open(datadir / 'results.csv', 'w') as csvfile:
        header = ['impl', 'procs', 'cons', 'size', 'time', 'latency', 'iops', 'ifutil']
        w = csv.DictWriter(csvfile, header)
        w.writeheader()
        for k in data:
            impl, procs, cons, size = k
            d = {'impl': impl, 'procs':procs, 'cons':cons, 'size': size}
            d.update(data[k])
            w.writerow(d)

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("datadir", help="directory containing the output of eval.py")

    args = parser.parse_args()
    calc_avgs(args.datadir)
