#!/usr/bin/env python3
import argparse
import asyncio
import multiprocessing
import os
import time

async def tcp_echo_connection(host, port, size, iterations):
    reader, writer = await asyncio.open_connection(host, port)

    out_data = bytearray(os.urandom(size))
    for i in range(1, iterations + 1):
        writer.write(out_data)

        in_data = await reader.read(size)
        assert(in_data == out_data)

    writer.close()

def client_main(args):
    connection_coroutines = []
    loop = asyncio.get_event_loop()

    start = time.time_ns()
    for _ in range(args.clients):
        connection_coro = tcp_echo_connection(args.addr, args.port, args.size, args.iterations)
        connection_coroutines.append(connection_coro)

    loop.run_until_complete(asyncio.gather(*connection_coroutines))
    end = time.time_ns()
    loop.close()

    if not args.quiet:
        print(f'{args.clients}; {args.iterations}; {args.size}; {end - start}')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Concurrent echo clients')
    parser.add_argument('-p', '--processes', type=int, default=1,
                        help='number of parallel client processes')
    parser.add_argument('-c', '--clients', type=int, default=10,
                        help='number of concurrent client connections per process')
    parser.add_argument('-i', '--iterations', type=int, default=1000,
                        help='number of echos each client connection sends')
    parser.add_argument('-a', '--addr', default='127.0.0.1',
                        help='address of the host')
    parser.add_argument('-P', '--port', type=int, default=12345,
                        help='port to connect to')
    parser.add_argument('-s', '--size', default=256, type=int,
                        help='size of the messages send to the server')
    parser.add_argument('-q', '--quiet', action='store_true',
                        help='suppress output')

    args = parser.parse_args()

    if not args.quiet:
        print('connections; iterations; nbyte; time[ns]')

    processes = []
    for _ in range(args.processes):
        process = multiprocessing.Process(target=client_main, args=(args,))
        processes.append(process)

    for process in processes:
        process.start()

    for process in processes:
        process.join()

