#include <assert.h>
#include <errno.h>
#include <netdb.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <sys/socket.h>

#include "liburing.h"

#define ECHOSIZE 32

static void* CONNECT_MARK = (void*)0;
static void* SEND_MARK = (void*)1;
static void* RECV_MARK = (void*)2;

static size_t cons;
static struct addrinfo* server;
static pthread_barrier_t readyBarrier;

static void die(char* msg) {
	perror(msg);
	exit(EXIT_FAILURE);
}

static void prepare_connect(struct io_uring* ring, int sock) {
    struct io_uring_sqe* sqe = io_uring_get_sqe(ring);
    assert(sqe);
    io_uring_prep_connect(sqe, sock, server->ai_addr, server->ai_addrlen);
    io_uring_sqe_set_data(sqe, CONNECT_MARK);
}

static void prepare_echo(struct io_uring* ring, int sock, char* in, char* out, size_t buf_len) {
    struct io_uring_sqe* sqe = io_uring_get_sqe(ring);
    assert(sqe);
    io_uring_prep_send(sqe, sock, in, buf_len, 0);
    io_uring_sqe_set_data(sqe, SEND_MARK);
    sqe->flags |= IOSQE_IO_LINK;

    sqe = io_uring_get_sqe(ring);
    assert(sqe);
    io_uring_prep_recv(sqe, sock, out, buf_len, 0);
    io_uring_sqe_set_data(sqe, RECV_MARK);
}

static void checked_io_uring_submit(struct io_uring* ring, int prepared) {
    int submitted = io_uring_submit(ring);
    if (submitted < 0) {
        errno = -submitted;
        die("io_uring_submit failed");
    }

    assert(submitted == prepared);
}

static void checked_io_uring_wait_cqe(struct io_uring* ring, struct io_uring_cqe** cqe_ptr) {
	int err = io_uring_wait_cqe(ring, cqe_ptr);
	if (err) {
		die("io_uring_wait_cqe failed");
	}
}

void* thread_func (__attribute__((unused)) void* arg) {
	struct io_uring ring;
	struct io_uring_cqe *cqe;

	int* sockets = malloc(sizeof(int) * cons);
	if (!sockets) {
		die("malloc failed");
	}

	if (io_uring_queue_init(8, &ring, 0) < 0) {
		die("io_uring_init_failed...");
	}

	for (size_t i = 0; i < cons; ++i) {
		sockets[i] = socket(server->ai_family, server->ai_socktype, server->ai_protocol);
		if (sockets[i] < 0) {
			die("socket failed");
		}

		prepare_connect(&ring, sockets[i]);
		checked_io_uring_submit(&ring, 1);
		checked_io_uring_wait_cqe(&ring, &cqe);

		assert(io_uring_cqe_get_data(cqe) == CONNECT_MARK);
		assert(cqe->res == 0);
		io_uring_cqe_seen(&ring, cqe);
	}

	pthread_barrier_wait(&readyBarrier);

	for (size_t i = 0; i < cons; ++i) {
		char in[ECHOSIZE];
		char out[ECHOSIZE];
		prepare_echo(&ring, sockets[i], &in[0], &out[0], ECHOSIZE);
		checked_io_uring_submit(&ring, 2);

		checked_io_uring_wait_cqe(&ring, &cqe);
		assert(io_uring_cqe_get_data(cqe) == SEND_MARK);
		assert(cqe->res == ECHOSIZE);
		io_uring_cqe_seen(&ring, cqe);

		checked_io_uring_wait_cqe(&ring, &cqe);
		assert(io_uring_cqe_get_data(cqe) == RECV_MARK);
		assert(cqe->res == ECHOSIZE);
		io_uring_cqe_seen(&ring, cqe);
	}

	for (size_t i = 0; i < cons; ++i) {
		shutdown(sockets[i], SHUT_RDWR);
		close(sockets[i]);
	}

	free(sockets);
	io_uring_queue_exit(&ring);

	return NULL;
}

int main(int argc, char* argv[]) {
	size_t nthreads;
	pthread_t* threads;
	if (argc != 5) {
		fprintf(stderr, "Usage: %s <addr> <port> <threads> <cons>\n", argv[0]);
		exit(EXIT_FAILURE); 
	}

	int err = getaddrinfo(argv[1], argv[2], NULL, &server);
	if (err) {
		if (err == EAI_SYSTEM) {                                                    
			die("getaddrinfo failed");
		} else {                                                                    
			fprintf(stderr, "error in getaddrinfo: %s\n", gai_strerror(err));
			exit(EXIT_FAILURE); 
		}
	}

	nthreads = strtol(argv[3], NULL, 10);
	cons = strtol(argv[4], NULL, 10);

	threads = malloc(sizeof(pthread_t) * cons);
	if (!threads) {
		die("malloc failed");
	}

	pthread_barrier_init(&readyBarrier, NULL, nthreads);

	for (size_t i = 0; i < nthreads; ++i) {
		pthread_create(&threads[i], NULL, thread_func, NULL);
	}

	for (size_t i = 0; i < nthreads; ++i) {
		pthread_join(threads[i], NULL);
	}

	return 0;
}
