#!/usr/bin/env python3
"""Evaluation script for the EMPER echoserver evaluation artifact"""

import argparse
import datetime
import fnmatch
import itertools
import logging
import os
from pathlib import Path
import subprocess
import sys
from time import sleep
from typing import Dict, List, Mapping, Sequence

from bench.benchresult import BenchResult
from bench.client import CLIENT_DIR, CLIENT_EXE, run_clients
from bench.globalvars import HOSTNAME, ROOT_DIR, TERMINATION_TIME
from bench.server import Server
from bench.server_cmds import SERVER_CMDS
from bench.util import cmd_run, prepare_env

ARTIFACT_DESC = subprocess.check_output(
    'git describe --dirty --always'.split(), cwd=ROOT_DIR, text=True)[:-1]

EMPER_20_WORKER_ENV_CMD = 'env EMPER_WORKER_COUNT=20'
EMPER_REPO = f'{ROOT_DIR}/emper'

EMPER_DEFAULT_CHECKOUT = 'origin/master'

CLIENT_CHECKOUT = EMPER_DEFAULT_CHECKOUT

EMPER_FIRST_CHECKOUT = 'muhq/io-first'

EMPER_LINKED_FUTURE_CHECKOUT = 'muhq/linked_future_echoserver'
EMPER_REAP_COMPLETIONS_PROVIDE_MEMORY_HEAP = 'muhq/reapCompletions_provide_memory_heap'

EMPER_WAKEUP_CONDVAR_CHECKOUT = 'cfba846c7763faf63877329b227c2b23af4fc568'

EMPER_BEFORE_COUNTING_TRY_LOCK_CHECKOUT = '5edc8e5eea4134badfa0bf1ee35bf8b671037328'
EMPER_BEFORE_BATCH_CQE_CHECKOUT = 'c8828c263618865866f20870c8488d8de11c776a'
EMPER_BEFORE_CHANGES_TO_WORKER_SLEEP_ALGO = 'cfba846c7763faf63877329b227c2b23af4fc568'

EMPER_BATCH_DEQUEUE_ANYWHERE = '6d63965bccb9c9612b85bbb46c31a1eacd3c6693'

EMPER_CQ_LOCK_CHECKOUT = 'muhq/cq_lock'

EMPER_BUF_SEL_CHECKOUT = 'muhq/buffer-selection'

EMPER_SQPOLL_CHECKOUT = 'muhq/sqpoll'

EMPER_FLAVORS = {
    'vanilla': {},
    #'buf-sel': {
    #    'checkout': EMPER_BUF_SEL_CHECKOUT,
    #},
    'pipe': {
        'meson_options': '-Dworker_sleep_strategy=pipe'
    },
    'pipe-no-completer': {
        'meson_options':
        '-Dworker_sleep_strategy=pipe -Dio_completer_behavior=none'
    },
    'no-stats': {
        'meson_options': '-Dstats=False'
    },
    'heap-alloc-continuation-buf': {
        'checkout': EMPER_REAP_COMPLETIONS_PROVIDE_MEMORY_HEAP
    },
    'before-batch-cqe': {
        'checkout': EMPER_BEFORE_BATCH_CQE_CHECKOUT,
        'meson_options': '-Dlog_level=OFF',
        'patches': ['fix_free_vs_delete.patch', 'before-batch-cqe.patch'],
    },
    'notify-all': {
        'meson_options': '-Dworker_wakeup_strategy=all'
    },
    'no-sleep': {
        'meson_options': '-Dworker_sleep=false'
    },
    'sqpoll': {
        'meson_options': '-Dio_uring_sq_poller=one',
        'checkout': EMPER_SQPOLL_CHECKOUT,
        'patches': ['getcpu.patch']
    },
    'sqpoll-numa': {
        'meson_options': '-Dio_uring_sq_poller=numa',
        'checkout': EMPER_SQPOLL_CHECKOUT,
        'patches': ['getcpu.patch']
    },
    'callback': {
        'patches': ['use_callback_echoserver.patch']
    },
    'wakeup-condvar': {
        'checkout': EMPER_WAKEUP_CONDVAR_CHECKOUT,
        'patches': ['fix_free_vs_delete.patch', 'wakeup-condvar.patch'],
    },
    'wakeup-all-condvar': {
        'checkout': EMPER_WAKEUP_CONDVAR_CHECKOUT,
        'meson_options': '-Dworker_wakeup_strategy=all',
        'patches': ['fix_free_vs_delete.patch', 'wakeup-condvar.patch'],
    },
    'wakeup-futex': {
        'meson_options': '-Dwakeup_semaphore_implementation=futex'
    },
    'wakeup-all-futex': {
        'meson_options':
        '-Dworker_wakeup_strategy=all -Dwakeup_semaphore_implementation=futex'
    },
    'shared-mutex': {
        'meson_option': '-Dlocked_unbounded_queue_implementation=shared_mutex'
    },
    'boost-shared-mutex': {
        'meson_option':
        '-Dlocked_unbounded_queue_implementation=boost_shared_mutex'
    },
    'pthread-rwlock': {
        'meson_option': '-Dlocked_unbounded_queue_implementation=rwlock'
    },
    'cq-mutex': {
        'meson_options': '-Dio_cq_lock_implementation=mutex'
    },
    'cq-spin-lock': {
        'meson_options': '-Dio_cq_lock_implementation=spin_lock'
    },
    'big-reap-cqe-section': {
        'checkout': EMPER_BEFORE_COUNTING_TRY_LOCK_CHECKOUT,
        'meson_options': '-Dlog_level=OFF',
        'patches': ['fix_free_vs_delete.patch', 'before-batch-cqe.patch'],
    },
    'completer-wakeup': {
        'meson_options': '-Dio_completer_behavior=maybe_wakeup'
    },
    'no-sleep-computation-100us': {
        'meson_options': '-Dworker_sleep=false',
        'server_cmd_suffix': '100'
    },
    'no-sleep-no-completer-computation-100us': {
        'meson_options':
        '-Dworker_sleep=false -Dio_completer_behavior=maybe_wakeup',
        'server_cmd_suffix': '100'
    },
    'no-sleep-computation-1ms': {
        'meson_options': '-Dworker_sleep=false',
        'server_cmd_suffix': ' 1000'
    },
    'no-sleep-no-completer-computation-1ms': {
        'meson_options':
        '-Dworker_sleep=false -Dio_completer_behavior=maybe_wakeup',
        'server_cmd_suffix': ' 1000'
    },
    'computation-100us': {
        'server_cmd_suffix': '100'
    },
    'computation-200us': {
        'server_cmd_suffix': ' 200'
    },
    'completer-wakeup-computation-100us': {
        'meson_options': '-Dio_completer_behavior=maybe_wakeup',
        'server_cmd_suffix': '100'
    },
    'computation-10us': {
        'server_cmd_suffix': '10'
    },
    'completer-wakeup-computation-10us': {
        'meson_options': '-Dio_completer_behavior=maybe_wakeup',
        'server_cmd_suffix': '10'
    },
    'mutex-computation-100us': {
        'meson_options': '-Dio_cq_lock_implementation=mutex',
        'server_cmd_suffix': '100'
    },
    'simple-arch': {
        'meson_options': '-Dio_single_uring=true',
    },
    'simple-arch-try-syscall': {
        'meson_options': '-Dio_single_uring=true -Dio_try_syscall=true',
    },
    'first': {
        'checkout': EMPER_FIRST_CHECKOUT,
        'patches': ['fix_free_vs_delete.patch', 'wakeup-condvar.patch'],
    },
    '4': {
        'server_cmd_prefix': 'env EMPER_WORKER_COUNT=4'
    },
    '20': {
        'server_cmd_prefix': EMPER_20_WORKER_ENV_CMD
    },
    '20-nopin': {
        'server_cmd_prefix':
        EMPER_20_WORKER_ENV_CMD + ' EMPER_PIN_WORKERS=false',
    },
    '20-computation-100us': {
        'server_cmd_prefix': EMPER_20_WORKER_ENV_CMD,
        'server_cmd_suffix': '100'
    },
    '20-nopin-computation-100us': {
        'server_cmd_prefix':
        EMPER_20_WORKER_ENV_CMD + ' EMPER_PIN_WORKERS=false',
        'server_cmd_suffix': '100'
    },
    '20-computation-200us': {
        'server_cmd_prefix': EMPER_20_WORKER_ENV_CMD,
        'server_cmd_suffix': '200'
    },
    '20-nopin-computation-200us': {
        'server_cmd_prefix':
        EMPER_20_WORKER_ENV_CMD + ' EMPER_PIN_WORKERS=false',
        'server_cmd_suffix': '200'
    },
    'linked-futures': {
        'checkout': EMPER_LINKED_FUTURE_CHECKOUT,
        'patches': ['fix_free_vs_delete.patch', 'wakeup-condvar.patch'],
    },
    'linked-futures-notify-all': {
        'checkout': EMPER_LINKED_FUTURE_CHECKOUT,
        'patches': ['fix_free_vs_delete.patch', 'wakeup-condvar.patch'],
        'meson_options': '-Dworker_wakeup_strategy=all'
    },
    'before-new-worker-sleep-algo': {
        'checkout': EMPER_BEFORE_CHANGES_TO_WORKER_SLEEP_ALGO,
        'patches': ['fix_free_vs_delete.patch', 'before-batch-cqe.patch'],
        'meson_options': '-Dlog_level=OFF',
    },
    'bdqfa': {
        'checkout': EMPER_BATCH_DEQUEUE_ANYWHERE,
    },
    'bdqfa-200us': {
        'checkout': EMPER_BATCH_DEQUEUE_ANYWHERE,
        'server_cmd_suffix': '200'
    },
    'bdqfa-20-nopin': {
        'checkout': EMPER_BATCH_DEQUEUE_ANYWHERE,
        'server_cmd_prefix':
        EMPER_20_WORKER_ENV_CMD + ' EMPER_PIN_WORKERS=false',
    },
    'bdqfa-20-nopin-200us': {
        'checkout': EMPER_BATCH_DEQUEUE_ANYWHERE,
        'server_cmd_prefix':
        EMPER_20_WORKER_ENV_CMD + ' EMPER_PIN_WORKERS=false',
        'server_cmd_suffix': '200'
    },
    'laws': {
        'meson_options':
        ('-Ddefault_scheduling_strategy=locality_aware_work_stealing '
         '-Dset_affinity_on_block=true'),
    },
    'laws-200us': {
        'meson_options':
        ('-Ddefault_scheduling_strategy=locality_aware_work_stealing '
         '-Dset_affinity_on_block=true'),
        'server_cmd_suffix':
        '200'
    },
    'laws-20-nopin': {
        'server_cmd_prefix':
        EMPER_20_WORKER_ENV_CMD + ' EMPER_PIN_WORKERS=false',
        'meson_options':
        ('-Ddefault_scheduling_strategy=locality_aware_work_stealing '
         '-Dset_affinity_on_block=true'),
    },
    'laws-20-nopin-200us': {
        'server_cmd_prefix':
        EMPER_20_WORKER_ENV_CMD + ' EMPER_PIN_WORKERS=false',
        'meson_options':
        ('-Ddefault_scheduling_strategy=locality_aware_work_stealing '
         '-Dset_affinity_on_block=true'),
        'server_cmd_suffix':
        '200'
    },
    'laws-bdqfa': {
        'checkout':
        EMPER_BATCH_DEQUEUE_ANYWHERE,
        'meson_options':
        ('-Ddefault_scheduling_strategy=locality_aware_work_stealing '
         '-Dset_affinity_on_block=true'),
    },
    'laws-bdqfa-200us': {
        'checkout':
        EMPER_BATCH_DEQUEUE_ANYWHERE,
        'meson_options':
        ('-Ddefault_scheduling_strategy=locality_aware_work_stealing '
         '-Dset_affinity_on_block=true'),
        'server_cmd_suffix':
        '200'
    },
    'laws-bdqfa-20-nopin': {
        'checkout':
        EMPER_BATCH_DEQUEUE_ANYWHERE,
        'server_cmd_prefix':
        EMPER_20_WORKER_ENV_CMD + ' EMPER_PIN_WORKERS=false',
        'meson_options':
        ('-Ddefault_scheduling_strategy=locality_aware_work_stealing '
         '-Dset_affinity_on_block=true'),
    },
    'laws-bdqfa-20-nopin-200us': {
        'checkout':
        EMPER_BATCH_DEQUEUE_ANYWHERE,
        'server_cmd_prefix':
        EMPER_20_WORKER_ENV_CMD + ' EMPER_PIN_WORKERS=false',
        'meson_options':
        ('-Ddefault_scheduling_strategy=locality_aware_work_stealing '
         '-Dset_affinity_on_block=true'),
        'server_cmd_suffix':
        '200'
    },
}

CONS = [500, 1000, 5000, 10000, 15000, 20000, 25000]
SIZES = [16]
# ARGS = itertools.product(CONS, SIZES)

MEASURE_CMD = "/usr/bin/time -o {BENCHDIR}/server_stats.{RUN}.txt -v"
MEASURE_IO_CMD = "sar 1 1000 -o {run}.{cons}.{size}.sar.data"
REMOTE_CMD = "ssh -p {ssh_port} -q {host}"
HOST = "faui49big02"

CLIENTS = None

SERVER_ENV = {}
CLIENT_ENV = {}

STARTUP_TIME = 10
CLIENT_SEPARATION_TIME = 10


def write_desc(data_dir):
    """Write a YAML description of the evaluation into result_dir"""
    desc_file_path = data_dir / 'desc.yml'
    with open(desc_file_path, 'w') as desc_file:
        print(f'start: {start_time.isoformat()}', file=desc_file)
        print(f'args: {ARGS}', file=desc_file)
        print(f'host: {HOST}', file=desc_file)
        print(f'measure_cmd: {MEASURE_CMD}', file=desc_file)
        print(f'measure_io_cmd: {MEASURE_IO_CMD}', file=desc_file)
        uname = os.uname()
        print(
            (f'uname_client: {uname.sysname} {uname.nodename} {uname.release} '
             f'{uname.version} {uname.machine}'),
            file=desc_file)

        if HOST == 'localhost':
            return

        # collect the uname of the host
        print('uname_host: ', end='', file=desc_file)

    # collect the uname of the host
    remote_cmd = REMOTE_CMD.format(host=HOST, ssh_port=SSH_PORT)
    uname_cmd = f'{remote_cmd} uname -a >> {desc_file_path}'
    cmd_run(uname_cmd, log)


def get_commit_id(repo=EMPER_REPO, commit_ish='HEAD') -> str:
    """Get git commit id of commitish from emper repo"""
    cmd = f'git rev-parse {commit_ish}'
    return subprocess.check_output(cmd.split(), cwd=repo, text=True)[:-1]


def emper_repo_add_muhq_remote():
    """Add muhq remote to emper repo"""
    cmd = 'git remote add muhq https://gitlab.cs.fau.de/aj46ezos/emper.git'
    subprocess.run(cmd.split(), cwd=EMPER_REPO, check=True)


def checkout_emper_worktree(directory, checkout):
    """Create a new worktree of checkout at directory"""
    worktree_cmd = f'git worktree add {directory} {checkout}'
    cmd_run(worktree_cmd, log, cwd=EMPER_REPO)


def build_emper(emper_dir, meson_options='', build_env=None):
    """Build a emper variant at emper_dir"""
    if not build_env:
        build_env = {}

    # add debug symbols to the release build
    build_env.update({'CFLAGS': '-g', 'CXXFLAGS': '-g'})
    meson_env = prepare_env(build_env)

    common_options = '-Dstats=true --buildtype=release --fatal-meson-warnings'
    build_cmd = f'meson build {common_options} {meson_options}'

    cmd_run(build_cmd, log, cwd=emper_dir, env=meson_env)

    ninja_cmd = 'ninja -C build'
    cmd_run(ninja_cmd, log, cwd=emper_dir)


def prepare_client():
    """Prepare and build the special emper variant for the client"""
    client_checkout = get_commit_id(EMPER_REPO, CLIENT_CHECKOUT)

    with open(DATA_DIR / 'desc.yml', 'a') as desc_file:
        print(f'client_commit: {client_checkout}', file=desc_file)
        # print(f'client_cmd: {CLIENT_CMD}', file=desc_file)

    if Path(f'{CLIENT_EXE}').exists():
        return

    print("Preparing the client ...")
    if not CLIENT_DIR.exists():
        checkout_emper_worktree(CLIENT_DIR, client_checkout)

    build_emper(CLIENT_DIR)


def update_emper_flavor(flavor_dir, checkout):
    """Reset and update a emper worktree"""
    # reset patches
    git_reset_cmd = 'git reset --hard'
    cmd_run(git_reset_cmd, log, cwd=flavor_dir)

    # update the worktree
    git_checkout_cmd = f'git checkout {checkout}'
    cmd_run(git_checkout_cmd, log, cwd=flavor_dir)

    # remove the old build directory
    rm_build_cmd = 'rm -r build'
    cmd_run(rm_build_cmd, log, cwd=flavor_dir)


def prepare_emper_flavors(flavors):
    """Prepare a emper flavor defined in EMPER_FLAVORS to be evaluated"""
    emper_server_cmd = SERVER_CMDS['emper']
    del SERVER_CMDS['emper']

    for flavor, desc in flavors.items():
        cmd_prefix = f'{desc["server_cmd_prefix"]} ' if 'server_cmd_prefix' in desc else ''
        cmd_suffix = f' {desc["server_cmd_suffix"]}' if 'server_cmd_suffix' in desc else ''
        cmd = emper_server_cmd.format(flavor=flavor)
        server_cmd = f'{cmd_prefix}{cmd}{cmd_suffix}'
        SERVER_CMDS[f'emper-{flavor}'] = server_cmd

        # write flavor description into the flavor's result dir
        flavor_dir = DATA_DIR / f'emper-{flavor}'
        if not flavor_dir.exists():
            os.mkdir(flavor_dir)

        checkout = get_commit_id(
            commit_ish=desc.get('checkout', EMPER_DEFAULT_CHECKOUT))
        meson_options = desc.get("meson_options", "")

        with open(flavor_dir / 'desc.yml', 'w') as flavor_desc_file:
            print(f'cmd: {server_cmd}', file=flavor_desc_file)
            print(f'commit: {checkout}', file=flavor_desc_file)
            print(f'meson_options: {meson_options}', file=flavor_desc_file)

        flavor_checkout = f'{ROOT_DIR}/emper-{flavor}'
        if Path(flavor_checkout).exists():
            if get_commit_id(flavor_checkout) != checkout:
                # old checkout
                print(f'updating {flavor_checkout}...')
                update_emper_flavor(flavor_checkout, checkout)
            else:
                # already correct checkout
                continue

        else:
            print(f'preparing {flavor_checkout}...')
            checkout_emper_worktree(flavor_checkout, checkout)

        for patch in desc.get('patches', []):
            patch_cmd = f'git apply {ROOT_DIR}/patches/{patch}'
            cmd_run(patch_cmd, log, cwd=flavor_checkout)

        build_emper(flavor_checkout,
                    meson_options=meson_options,
                    build_env=desc.get('build_env', {}))


def summarize(results: Mapping[str, Sequence[Sequence[BenchResult]]]):
    """Create, print and safe a summary of the evaluation"""
    summary = ''
    totals = {s: 0 for s in BenchResult}
    for server, server_results in results.items():
        for bench_results in server_results:
            for i, arg in enumerate(ARGS):
                summary += f'{i + 1}/{len(ARGS)} {bench_results[i].name} {server} / {arg}\n'
                totals[bench_results[i]] += 1

    summary += '\n'
    for result, amount in totals.items():
        summary += f'{result.name}: {amount}\n'

    print(summary)
    with open(DATA_DIR / 'summary.txt', 'w') as summary_file:
        print(summary, file=summary_file)


RunResult = List[BenchResult]
ServerResult = List[RunResult]
EvalResult = Dict[str, ServerResult]


def bench() -> EvalResult:
    """Run the benchmark for all selected servers and client arguments"""
    results = {
        server: [[BenchResult.SKIP] * len(ARGS) for r in range(RUNS)]
        for server in SERVER_CMDS
    }

    for run in range(1, RUNS + 1):
        print(f"starting run {run}/{RUNS}...\u001b[K")

        for server_name, server_cmd in SERVER_CMDS.items():
            bench_results = results[server_name][run - 1]

            bench_dir = DATA_DIR / server_name
            if not bench_dir.exists():
                os.mkdir(bench_dir)

            substitutions = {
                'BENCHDIR': bench_dir,
                'RUN': run,
                'SERVER': server_name
            }

            remote_cmd = None
            if REMOTE_CMD:
                remote_cmd = REMOTE_CMD.format(host=HOST, ssh_port=SSH_PORT)
                # make sure the benchdir is available on the HOST
                remote_prepare_cmd = f'{remote_cmd} mkdir -p {bench_dir}'
                cmd_run(remote_prepare_cmd, log)

            measure_cmd = MEASURE_CMD.format(**substitutions)

            print(f'benchmarking {server_name} ...')
            server = Server(server_name,
                            server_cmd,
                            run,
                            bench_dir,
                            env=SERVER_ENV,
                            measure_cmd=measure_cmd,
                            remote_cmd=remote_cmd,
                            host=HOST)

            sleep(STARTUP_TIME)
            if server.poll() is not None:
                log.error('server cmd returned early')
                continue

            for i, (cons, size) in enumerate(ARGS):
                # clear line to the right and reset cursor to the first column
                print(f'{i + 1}. (c {cons} s {size}) of {len(ARGS)}\u001b[K\r',
                      end='')

                # skip naive-multithreaded for cons > 10000
                if server_name == 'naive-multithreaded' and cons > 10000:
                    log.debug(
                        'Skipping naive-multithreaded for cons (%d) > 10000',
                        cons)
                    continue

                measure_io_proc = None
                if MEASURE_IO_CMD:
                    measure_io_cmd = MEASURE_IO_CMD.format(cons=cons,
                                                           size=size,
                                                           run=run)

                    measure_io_proc = subprocess.Popen(
                        measure_io_cmd.split(),
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                        cwd=bench_dir)

                client_args = {'cons': cons, 'size': size}
                bench_results[i] = run_clients(run,
                                               client_args,
                                               bench_dir,
                                               HOST,
                                               substitutions=substitutions,
                                               termination=TERMINATION,
                                               remote_cmd=remote_cmd,
                                               hosts=CLIENTS,
                                               env=CLIENT_ENV)

                if bench_results[i] == BenchResult.FAIL:
                    log.warning('\nClient cmd failed. Terminate server')
                    break

                # stop io stats collection
                if measure_io_proc:
                    measure_io_proc.kill()
                    measure_io_proc.wait(TERMINATION_TIME)

                sleep(CLIENT_SEPARATION_TIME)

            server.shutdown()
    return results


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-r',
                        '--runs',
                        help='benchmark runs',
                        type=int,
                        default=3)
    parser.add_argument('-c',
                        '--connections',
                        help='number of client connections to evaluate',
                        type=int,
                        nargs='+')
    parser.add_argument('--clients',
                        help='client computers used to evaluate the host',
                        type=str,
                        nargs='+')
    parser.add_argument('-s',
                        '--sizes',
                        help='sizes of messages sent',
                        type=int,
                        nargs='+')
    parser.add_argument('-f',
                        '--flavors',
                        help='emper flavors to benchmark',
                        nargs='+')
    parser.add_argument('-m',
                        '--measure-cmd',
                        help='command executing the server',
                        default=MEASURE_CMD)
    parser.add_argument('--measure-io',
                        help='measure the IO done by all network interfaces',
                        default=False,
                        action='store_true')
    parser.add_argument('--ssh-port',
                        help='ssh port to use for the connection to the host',
                        default=22,
                        type=int)
    parser.add_argument(
        '--data-root',
        help='path where the experiment results should be saved',
        type=str)
    parser.add_argument('--log', help='Log level to use', type=str)
    parser.add_argument(
        '--additional-client-args',
        help='additional arguments appended to the echo client command',
        type=str)
    parser.add_argument('-nf',
                        '--no-fetch',
                        help='do not fetch the emper remotes',
                        action='store_true')
    parser.add_argument('implementations',
                        help='server implementations to benchmark',
                        nargs='*')

    location_group = parser.add_mutually_exclusive_group()
    location_group.add_argument('-l',
                                '--local',
                                help='run benchmark on localhost',
                                action='store_true')
    location_group.add_argument('--host', help='host running the echo servers')

    termination_group = parser.add_mutually_exclusive_group()
    termination_group.add_argument(
        '--time', help='time the echo client should issue echos', type=int)
    termination_group.add_argument(
        '--iterations',
        help='echos the client should issue per connection',
        type=int)

    args = parser.parse_args()

    if args.log:
        numeric_loglevel = getattr(logging, args.log.upper(), None)
        if not isinstance(numeric_loglevel, int):
            raise ValueError(f'Invalid log level: {args.log}')
    else:
        numeric_loglevel = logging.WARNING
    logging.basicConfig(level=numeric_loglevel)
    log = logging.getLogger(Path(__file__).name)

    TERMINATION = None
    if args.time:
        TERMINATION = ('time', args.time)
    elif args.iterations:
        TERMINATION = ('iterations', args.iterations)

    SSH_PORT = args.ssh_port

    CLIENTS = args.clients

    if not Path(EMPER_REPO).exists():
        subprocess.run('git subproject update --init'.split(), check=True)
        emper_repo_add_muhq_remote()
    else:
        if not args.no_fetch:
            # fetch the emper repository to be sure we have the most recent version
            emper_fetch_cmd = 'git fetch --all'.split()
            subprocess.run(emper_fetch_cmd, cwd=EMPER_REPO, check=True)

    if not args.measure_io:
        MEASURE_IO_CMD = ''

    if args.local:
        # split the available core between both emper processes
        ncpus = os.cpu_count()
        if ncpus is not None:
            workerCount = ncpus // 2
            SERVER_ENV['EMPER_WORKER_COUNT'] = str(workerCount)
            CLIENT_ENV['EMPER_WORKER_COUNT'] = str(workerCount)
            CLIENT_ENV['EMPER_PINNING_OFFSET'] = str(workerCount - 1)
        else:
            log.warning(('can not determine cpu count and worker count.'
                         ' Each EMPER process will use all available cpus'))

        REMOTE_CMD = ""

        HOST = "localhost"

    if args.host:
        HOST = args.host

    if args.implementations:
        SERVER_CMDS = {
            i: c
            for i, c in SERVER_CMDS.items() if i in args.implementations
        }

    MEASURE_CMD = args.measure_cmd

    _cons = args.connections or CONS
    _sizes = args.sizes or SIZES
    ARGS = list(itertools.product(_cons, _sizes))

    # select emper flavors to evaluate
    selected_flavors = EMPER_FLAVORS
    if args.flavors:
        filtered_flavors = {}
        for selected_flavor in args.flavors:
            selection = {
                f: selected_flavors[f]
                for f in fnmatch.filter(selected_flavors.keys(),
                                        selected_flavor)
            }
            if not selection:
                log.critical('%s does not select any available flavor',
                             selected_flavor)
                sys.exit(1)

            filtered_flavors.update(selection)

        selected_flavors = filtered_flavors

    print("Building the artifact ...")
    make_cmd = ['make']
    if numeric_loglevel > logging.DEBUG:
        make_cmd.append('-s')
    subprocess.run(make_cmd, check=True, cwd=ROOT_DIR)

    start_time = datetime.datetime.now()
    experiment_desc = f'{ARTIFACT_DESC}-{HOSTNAME}-{HOST}'

    if args.data_root:
        data_root = Path(args.data_root)
    else:
        data_root = ROOT_DIR / "results"
    DATA_DIR = data_root / experiment_desc / f'{start_time.strftime("%Y-%m-%dT%H_%M_%S")}'

    if not DATA_DIR.exists():
        os.makedirs(DATA_DIR)

    print(f"Storing results at {DATA_DIR}")

    write_desc(DATA_DIR)
    prepare_client()

    # generate emper flavors
    if 'emper' in SERVER_CMDS:
        prepare_emper_flavors(selected_flavors)

    RUNS = args.runs

    if not RUNS > 0:
        sys.exit(0)

    print()
    summarize(bench())
