#!/usr/bin/env python3

import argparse
from configparser import ConfigParser
import csv
import fnmatch
import io
from pathlib import Path
from typing import Iterable, Mapping, Optional, Sequence, Union

import numpy as np

Number = Union[int, float]


def to_number(string: str) -> Number:
    # we convert always to float so python parses scientific notation
    f = float(string)
    return f if '.' in string else int(f)


DataPoint = Union[list[Number], Number]
ExperimentData = dict[str, DataPoint]
ImplData = dict[int, list[ExperimentData]]
EvaluationResults = dict[str, ImplData]


def collect_results(result_dir: Union[Path, str],
                    include: Sequence[str] = None,
                    exclude: Sequence[str] = None,
                    file_extension="ini",
                    print_experiments=False) -> EvaluationResults:
    result_dir = Path(result_dir)
    impl_dirs = {d.name: d for d in result_dir.iterdir() if d.is_dir()}

    if include:
        filtered_impl_dirs = {}
        for impl in include:
            filtered_impl_dirs.update({
                i: impl_dirs[i]
                for i in fnmatch.filter(impl_dirs.keys(), impl)
            })
        impl_dirs = filtered_impl_dirs

    if exclude:
        filtered_impl_dirs = {}
        for pattern in exclude:
            filtered_impl_dirs.update({
                i: impl_dirs[i]
                for i in impl_dirs
                if i not in fnmatch.filter(impl_dirs.keys(), pattern)
            })
        impl_dirs = filtered_impl_dirs

    data = {}

    for impl, impl_dir in impl_dirs.items():
        impl_d: ImplData = {}

        data_files = impl_dir.glob(f'*.{file_extension}')

        # collect all data
        for data_path in data_files:
            configparser = ConfigParser()
            configparser.read(data_path)

            exp_data: ExperimentData = {}
            error = False
            for k in ['connect_duration', 'echo_duration', 'total_duration']:
                try:
                    exp_data[k] = to_number(configparser['global'][k])
                except KeyError:
                    error = True
                    print(f'ERROR in {data_path}')
                    break

            if error:
                continue

            csv_data = []
            csvfile = io.StringIO(configparser['clients']['csv'][1:])
            reader = csv.DictReader(csvfile)

            # convert everything to numbers if possible
            for row in list(reader):
                csv_data.append({k: to_number(v) for k, v in row.items()})

            cons = int(configparser['global']['clients'])

            client_iterations = [c['iterations'] for c in csv_data]
            exp_data['client-iterations'] = client_iterations

            echos = sum(client_iterations)
            exp_data['total_iterations'] = echos
            # We know that echo_duration is a Number so ignore mypy
            iops = echos / (exp_data['echo_duration'] * 10**-9)  # type: ignore
            exp_data['iops'] = iops

            if cons in impl_d:
                impl_d[cons] += [exp_data]
            else:
                impl_d[cons] = [exp_data]

            if print_experiments:
                print(
                    f'{data_path}: clients: {configparser["global"]["clients"]} ',
                    end='')
                print(exp_data)

        # only add the implementations data if there is actual data
        if impl_d:
            data[impl] = impl_d

    return data


Stats = dict[str, float]
ConnectionStats = dict[str, Stats]
ImplStats = dict[int, ConnectionStats]
EvaluationStats = dict[str, ImplStats]


def calculate_stats(data: EvaluationResults, warn=False) -> EvaluationStats:
    stats = {}
    for impl, impl_d in data.items():
        impl_stats = {}

        for cons in impl_d:
            cons_stats = {}

            for key in impl_d[cons][0]:
                _values = [x[key] for x in impl_d[cons]]

                if isinstance(_values[0], list):
                    values = [x for l in _values for x in l]  # type: ignore
                else:
                    values = _values

                cur_stats = {}
                cur_stats['mean'] = np.mean(values)
                cur_stats['median'] = np.median(values)
                cur_stats['min'] = min(values)
                cur_stats['max'] = max(values)
                cur_stats['std'] = np.std(values)

                if warn:
                    precision = cur_stats['std'] / cur_stats['mean']
                    if precision > 0.05:
                        print('Warning: imprecise data impl:', end='')
                        print(
                            f' {impl}, c: {cons}, k: {key} p: {precision*100:.5}%'
                        )

                cons_stats[key] = cur_stats

            impl_stats[cons] = cons_stats

        stats[impl] = impl_stats

    return stats


def print_stats_for_variable(variable: str, stats: Mapping):
    print(f'{variable}: ', end='')
    for name, value in stats.items():
        print(f'{name}: {value}, ', end='')
    print(f'std[%]: {(stats["std"]/stats["mean"])*100:.2f}%')


def print_stats(stats: EvaluationStats, variables: Optional[Iterable]):
    for impl, impl_stats in stats.items():
        print(f'{impl}:')
        for cons in sorted(list(impl_stats.keys())):
            cons_stats = impl_stats[cons]
            print(f'   {cons} connections:')

            if not variables:
                variables = cons_stats.keys()

            for variable in variables:
                print('    ', end='')
                print_stats_for_variable(variable, cons_stats[variable])


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("result_dir")
    parser.add_argument("-w",
                        "--warn",
                        help='warn about imprecise data',
                        action='store_true')
    parser.add_argument("--print-experiments",
                        help='print results from each individual experiment',
                        action='store_true')
    parser.add_argument("-i",
                        "--implementations",
                        help="implementations to collect",
                        nargs='+')
    parser.add_argument("-ix",
                        "--exclude-implementations",
                        help="implementations to exclude",
                        nargs='+')
    parser.add_argument(
        "--print-stats",
        help="variables about which the collected statistics should be print",
        nargs='*',
        default=['iops'])

    args = parser.parse_args()

    data = collect_results(args.result_dir,
                           include=args.implementations,
                           exclude=args.exclude_implementations,
                           print_experiments=args.print_experiments)

    stats = calculate_stats(data, args.warn)

    if args.print_stats is not None:
        print_stats(stats, variables=args.print_stats)


if __name__ == '__main__':
    main()
