#!/usr/bin/env python3

import argparse
import ast
import operator
from string import Template
from typing import cast

import parse_results
import plot_utils


#https://stackoverflow.com/questions/2371436/evaluating-a-mathematical-expression-in-a-string
def _eval_with_stat(evaluation: str, stats, stat: str):
    """Helper to evaluate a datapoint description string as arithmetic operation"""
    def _eval(node):
        """evaluate a arithmetic ast node"""

        # supported operators
        operators = {
            ast.Add: operator.add,
            ast.Sub: operator.sub,
            ast.Mult: operator.mul,
            ast.Div: operator.truediv,
            ast.Pow: operator.pow,
            ast.BitXor: operator.xor,
            ast.USub: operator.neg
        }

        if isinstance(node, ast.Num):  # <number>
            return node.n
        if isinstance(node, ast.BinOp):  # <left> <operator> <right>
            return operators[type(node.op)](_eval(node.left),
                                            _eval(node.right))
        if isinstance(node, ast.UnaryOp):  # <operator> <operand> e.g., -1
            return operators[type(node.op)](_eval(node.operand))

        raise TypeError(node)

    if evaluation in stats:
        return stats[evaluation][stat]

    expr = evaluation.format(**{k: v[stat] for k, v in stats.items()})

    node = cast(ast.Expression, ast.parse(expr, mode='eval'))

    return _eval(node.body)


def build_tikz(stats,
               y_expression,
               output_path,
               result_dir,
               y_label=None,
               error='std'):
    if y_label is None:
        y_label = y_expression

    mapping = {}
    preambel = \
"""\\documentclass{standalone}
\\input{common.tex}

\\begin{document}

\\begin{tikzpicture}
\\begin{groupplot}[
        group style={
            group name=plots,
            group size=1 by 1,
            vertical sep=0.65cm,
            horizontal sep=0.7cm,
            xticklabels at=edge bottom,
            xlabels at=edge bottom,
            ylabels at=edge left,
        },
        cycle list name=vibrant,
        xtick pos=bottom,
        ytick pos=left,
        xminorticks=false,
        yminorticks=false,
        ymajorgrids=true,
        height=7.5cm,
        width=10cm,
        enlarge x limits={
            auto,
            value=0.04,
        },
        legend columns=3,
        legend style={
            at={(.5,1.05)},
            anchor=south,
            font=\\scriptsize,
            draw=none,
            fill=none,
        },
]

"""

    nextgroupplot = Template("""
\\nextgroupplot[xtick={$x}, xticklabels={$xlabels}]
""")

    addplot = Template("""
\\addplot+[sharp plot, error bars/.cd, y dir=both, y explicit]
    table [x=x, y=y, $error_def]{
x\ty\t$error_names
$table};
""")

    addlegendentry = Template("""
\\addlegendentry{$impl}
""")

    tail = Template("""
\\end{groupplot}

\\draw ($$(plots c1r1.south) + (0,-0.4)$$) node[anchor=north] (xlabel) {\\textbf{Connections}};
\\draw ($$(plots c1r1.west) + (-0.8,0)$$) node[anchor=south,rotate=90] (ylabel) {\\textbf{$y_label}};

\\node (title) at ($$(plots c1r1.south) - (0,1.2)$$) {$title};

\\end{tikzpicture}
\\end{document}
""")

    longest_cons = []
    for impl, impl_stats in stats.items():
        all_cons = list(impl_stats)
        all_cons.sort()
        if not longest_cons or len(all_cons) > len(longest_cons):
            longest_cons = all_cons

        impl_map_name = impl.replace('_', '-')
        c_map = ''
        for i, cons in enumerate(all_cons):
            y_val = _eval_with_stat(y_expression, impl_stats[cons], 'median')

            data_row = f'{i}\t{y_val}\t'

            if error == 'std':
                std = _eval_with_stat(y_expression, impl_stats[cons], 'std')
                data_row += f'{std}'
            elif error == 'minmax':
                _min = _eval_with_stat(y_expression, impl_stats[cons], 'min')
                _max = _eval_with_stat(y_expression, impl_stats[cons], 'max')
                data_row += f'{y_val - _min}\t{_max -y_val}'

            c_map += f'{data_row}\n'

        mapping[impl_map_name] = c_map

    tikz = preambel
    x_vals = ','.join([str(x) for x in range(len(longest_cons))])
    xlabels = ','.join([str(x) for x in longest_cons])
    tikz += nextgroupplot.substitute(x=x_vals, xlabels=xlabels)

    ordered_impls = list(mapping.keys())
    ordered_impls.sort()

    assert (error in ['std', 'minmax'])
    if error == 'std':
        error_def = 'y error=std'
        error_names = 'std'
    elif error == 'minmax':
        error_def = 'y error minus=ey-, y error plus=ey+'
        error_names = 'ey-\tey+'

    for impl in ordered_impls:
        table = mapping[impl]
        tikz += addplot.substitute(table=table,
                                   error_def=error_def,
                                   error_names=error_names)
        impl_label = plot_utils.subst_runtime_name(impl)
        tikz += addlegendentry.substitute(impl=impl_label)
    tikz += tail.substitute(y_label=y_label,
                            title=result_dir.replace('_', '\\_'))

    with open(output_path, 'w') as output_file:
        print(tikz, file=output_file)


def plot_matplotlib(stats, y_expression, result_dir):  # pylint: disable=unused-argument
    pass


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("result_dir")
    parser.add_argument("-w",
                        "--warn",
                        help='warn about imprecise data',
                        action='store_true')
    parser.add_argument("-o", "--outfile", default="echo.tex")
    parser.add_argument("-i",
                        "--implementations",
                        help="imnplementations to plot",
                        nargs='+')
    parser.add_argument("-ix",
                        "--exclude-implementations",
                        help="imnplementations to exclude",
                        nargs='+')
    parser.add_argument("--error-bars",
                        default='std',
                        choices=['std', 'minmax'],
                        help="plot format to generate")
    parser.add_argument("-f",
                        "--format",
                        default="tikz",
                        choices=['tikz', 'matplotlib'],
                        help="plot format to generate")
    parser.add_argument("-y",
                        "--y-expression",
                        default="{iops}/1000",
                        help="expression to evaluate as y value")
    parser.add_argument("--y-label", help="label of the y axis")

    args = parser.parse_args()

    data = parse_results.collect_results(args.result_dir,
                                         include=args.implementations,
                                         exclude=args.exclude_implementations)
    stats = parse_results.calculate_stats(data, args.warn)

    if args.format == 'tikz':
        build_tikz(stats,
                   args.y_expression,
                   args.outfile,
                   args.result_dir,
                   y_label=args.y_label,
                   error=args.error_bars)
    elif args.format == 'matplotlib':
        plot_matplotlib(stats, args.y_expression, args.result_dir)


if __name__ == '__main__':
    main()
