#!/usr/bin/env python3
import argparse
import pathlib
from string import Template

import numpy as np
from matplotlib import pyplot as plt

import plot_utils

CSV_FIELDS = {'total': 0, 'after-send': 1, 'after-send-dispatch': 2}


def get_bar_data(latency_files, latency):

    field = CSV_FIELDS[latency]
    prctl = [50, 95, 99, 99.99, 99.999]

    data = {}
    for latency_file in latency_files:
        path = pathlib.Path(latency_file)
        impl = path.parent.name
        ns = np.sort(
            np.loadtxt(path,
                       dtype=int,
                       delimiter=',',
                       skiprows=1,
                       usecols=field))

        mean = np.mean(ns)
        median, p95, p99, p99_99, p99_999 = np.percentile(ns, prctl)
        data[impl] = {
            'mean': mean,
            'median': median,
            95: p95,
            99: p99,
            99.99: p99_99,
            99.999: p99_999
        }

    return data


def plot_tail_latency_bar_tikz(latency_files, latency):
    preamble = Template("""\\documentclass{standalone}
\\input{common.tex}

\\begin{document}

% data: $latency_files
\\begin{tikzpicture}
\\begin{axis}[
    ybar,
    enlargelimits=0.15,
    legend style={
        at={(0.5,-0.15)},
        anchor=north,
        legend columns=-1},
    ylabel={latency in ms},
    xlabel={percentiles},
    cycle list name=barvibrant,
    symbolic x coords={95\\%,99\\%,99.99\\%,99.999\\%},
    xtick=data,
    nodes near coords,
    nodes near coords align={horizontal},
    nodes near coords style={rotate=90},
]
""")

    addplot = Template(
        "\\addplot coordinates {(95\\%, $p95) (99\\%, $p99) (99.99\\%, $p99_99) (99.999\\%, $p99_999)};"
    )

    tail = Template("""\\legend{$labels}
\\end{axis}
\\end{tikzpicture}
\\end{document}
""")

    labels = ''
    plots = []
    for impl, data in get_bar_data(latency_files, latency).items():
        impl_label = plot_utils.subst_runtime_name(impl)
        labels += f'{impl_label}, '
        # convert to ms
        p95, p99, p99_99, p99_999 = np.array([
            data[95],
            data[99],
            data[99.99],
            data[99.999],
        ]) // 1000000

        plots.append(
            addplot.substitute(p95=p95,
                               p99=p99,
                               p99_99=p99_99,
                               p99_999=p99_999))

    print(preamble.substitute(latency_files=latency_files))
    for plot in plots:
        print(plot)
    print(tail.substitute(labels=labels[:-2]))


def plot_tail_latency_bar(latency_files, latency):
    n_impls = len(latency_files)
    width = 1 / (n_impls + 1)

    impl_bars = []
    for i, (impl,
            data) in enumerate(get_bar_data(latency_files, latency).items()):

        x_base = np.arange(1, len(data) + 1)
        x = x_base + width / 2 + (i * width)
        y = np.array([
            data['mean'], data['median'], data[95], data[99], data[99.99],
            data[99.999]
        ])
        # convert to ms
        y = y / 1000000
        print(x, y)
        impl_label = plot_utils.subst_runtime_name(impl)
        impl_bars.append(plt.bar(x, y, label=impl_label, width=width))

    # Add counts above the two bar graphs
    for bar in impl_bars:
        for rect in bar:
            height = rect.get_height()
            plt.text(rect.get_x() + rect.get_width() / 2,
                     height,
                     f'{int(height)}',
                     ha='center',
                     va='bottom',
                     rotation=45)

    plt.ylabel(f'{latency} latency in ms')
    plt.xlabel('percentiles')
    plt.xticks(x_base + (n_impls / 2 * width),
               ['mean', 'median', '95%', '99%', '99.99%', '99.999%'])
    plt.legend()
    plt.show()


def plot_tail_latency_linear(latency_files, latency):
    field = CSV_FIELDS[args.latency]
    for latency_file in latency_files:
        path = pathlib.Path(latency_file)
        ns = np.sort(
            np.loadtxt(path,
                       dtype=int,
                       delimiter=',',
                       skiprows=1,
                       usecols=field))

        echos = len(ns)
        prct_90 = int(echos * 0.9)
        in_data = echos - prct_90
        ms = ns[prct_90:] // 1000000

        plt.plot(ms, label=path.parent.name)

    plt.xticks([0, in_data // 2, in_data], ['90%', '95%', '100%'])
    plt.ylabel(f'{latency} latency in ms')
    plt.legend()
    plt.show()


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("data_files", nargs='+')
    parser.add_argument("--latency",
                        choices=['total', 'after-send', 'after-send-dispatch'],
                        default='total')
    parser.add_argument("--plot",
                        choices=['line', 'bar', 'bar-tikz'],
                        default='line')

    args = parser.parse_args()

    if args.plot == 'bar':
        plot_tail_latency_bar(args.data_files, args.latency)
    elif args.plot == 'line':
        plot_tail_latency_linear(args.data_files, args.latency)
    else:
        plot_tail_latency_bar_tikz(args.data_files, args.latency)
