/*
 * echoserver.c
 *
 * This is a modified version of the "simpler ROT13 server with
 * Libevent" from:
 * http://www.wangafu.net/~nickm/libevent-book/01_intro.html
 */

#include <err.h>
#include <errno.h>
#include <event2/buffer.h>
#include <event2/bufferevent.h>
#include <event2/event.h>
#include <event.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>

#define PORT 12345
#define BACKLOG 10240

#define unlikely(x) __builtin_expect(!!(x), 0)

#define xstr(a) str(a)
#define str(a) #a

struct client {
  /* The clients socket. */
  int fd;
};

void on_read(struct bufferevent *bev, void *ctx) {
  struct evbuffer *input;
  struct evbuffer_ptr p;

  input = bufferevent_get_input(bev);

  evbuffer_ptr_set(input, &p, 0, EVBUFFER_PTR_SET);
  p = evbuffer_search(input, "quit\n", 5, &p);

  if (p.pos != -1) {
    exit(EXIT_SUCCESS);
  }

  int error = bufferevent_write_buffer(bev, input);
  if (unlikely(error)) {
    err(error, "writing input to output failed:");
    bufferevent_free(bev);
  }
}

void on_error(struct bufferevent *bev, short error, void *ctx) {
  struct client *client = (struct client *)ctx;

  if (!(error & EVBUFFER_EOF)) {
    err(error, "Client disconnecting");
  }

  close(client->fd);
  free(client);
  bufferevent_free(bev);
}

void on_accept(evutil_socket_t listener, short event, void *arg) {
  struct event_base *base = arg;
  int fd = accept(listener, NULL, NULL);
  if (fd < 0) {
    perror("accept failed");
    return;
  }

  struct client* client = malloc(sizeof(client));
  if (unlikely(!client)) {
    err(EXIT_FAILURE, "malloc failed");
  }
  client->fd = fd;


  struct bufferevent *bev;
  evutil_make_socket_nonblocking(fd);
  bev = bufferevent_socket_new(base, fd, BEV_OPT_CLOSE_ON_FREE);
  bufferevent_setcb(bev, on_read, NULL, on_error, client);
  bufferevent_enable(bev, EV_READ);
}

int main(int argc, char **argv) {
  setvbuf(stdout, NULL, _IONBF, 0);

  evutil_socket_t listener;
  struct addrinfo* servaddr;
  struct event_base *base;
  struct event *listener_event;

  base = event_base_new();
  if (!base) return 1;

  int error = getaddrinfo("::", xstr(PORT), NULL, &servaddr);
  if (error) {
    if (error) {
      err(EXIT_FAILURE, "getaddrinfo failed");
    } else {
      fprintf(stderr, "error in getaddrinfo: %s\n", gai_strerror(error));
      return EXIT_FAILURE;
    }
  }

  listener = socket(servaddr->ai_family, servaddr->ai_socktype, servaddr->ai_protocol);
  if (listener < 0) {
    err(EXIT_FAILURE, "listen failed");
  }

  int reuseaddr_on = 1;
  setsockopt(listener, SOL_SOCKET, SO_REUSEADDR, &reuseaddr_on, sizeof(reuseaddr_on));

  evutil_make_socket_nonblocking(listener);

  if (bind(listener, servaddr->ai_addr, servaddr->ai_addrlen) < 0) {
    err(EXIT_FAILURE, "bind failed");
  }

  if (listen(listener, BACKLOG) < 0) {
    err(EXIT_FAILURE, "listen failed");
  }

  listener_event =
      event_new(base, listener, EV_READ | EV_PERSIST, on_accept, (void *)base);

  event_add(listener_event, NULL);

  event_base_dispatch(base);

  return 0;
}
