// SPDX-License-Identifier: GPL-3.0-or-later
// Copyright © 2020 Florian Fischer
/* #include <assert.h> */
#include <errno.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>

#define PORT 12345
#define BUF_MAX 1024
#define BACKLOG 10240

#define unlikely(x) __builtin_expect(!!(x), 1)

#define xstr(a) str(a)
#define str(a) #a

_Noreturn void die(const char* msg) {
	perror(msg);
	exit(EXIT_FAILURE);
}

ssize_t send_all(int fd, const void* buf, size_t len) {
	size_t bytes_written = 0;
	while (bytes_written < len) {
		ssize_t new_bytes_written = send(fd, buf, len - bytes_written, MSG_NOSIGNAL);
		if (unlikely(new_bytes_written < 1)) {
			return new_bytes_written;
		}

		bytes_written += new_bytes_written;
	}

	return (ssize_t)bytes_written;
}

void* client_func(void* arg) {
	int client_fd = *(int*)arg;
	free(arg);

	char buf[BUF_MAX];
	for (;;) {
		ssize_t bytes_recv = recv(client_fd, (void*)&buf, BUF_MAX, 0);
		if (unlikely(bytes_recv <= 0)) {
			if (unlikely(bytes_recv == -1)) {
				perror("recv failed");
			}
			goto finish;
		}

		if (bytes_recv == 5 && strncmp(buf, "quit\n", 5) == 0) {
			exit(EXIT_SUCCESS);
		}

		ssize_t bytes_send = send_all(client_fd, buf, bytes_recv);
		if (unlikely(bytes_send < 0)) {
			perror("send failed");
			goto finish;
		}
	}

finish:
	close(client_fd);
	return NULL;
}

int main(int argc, char* argv[]) {
	struct addrinfo* servaddr;
	int err = getaddrinfo("::", xstr(PORT), NULL, &servaddr);
	if (err) {
		if (err == EAI_SYSTEM) {
			die("getaddrinfo failed");
		} else {
			fprintf(stderr, "error in getaddrinfo: %s\n", gai_strerror(err));
			return EXIT_FAILURE;
		}
	}

	int sockfd = socket(servaddr->ai_family, servaddr->ai_socktype, servaddr->ai_protocol);
	if (sockfd == -1) {
		die("socket failed");
	}

	int reuseaddr = 1;
	if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &reuseaddr, sizeof(reuseaddr)) == -1) {
		die("setsockopt failed");
	}

	if (bind(sockfd, servaddr->ai_addr,servaddr->ai_addrlen) == -1) {
		die("bind failed");
	}

	if (listen(sockfd, BACKLOG) != 0) {
		die("listen failed");
	}

	for (;;) {
		int client_fd = accept(sockfd, NULL, NULL);
		if (unlikely(client_fd < 0)) {
			die("accept failed");
		}

		int* passable_fd = malloc(sizeof(int));
		if (unlikely(!passable_fd)) {
			die("allocating passable fd failed");
		}

		*passable_fd = client_fd;

		pthread_t thread;
		pthread_create(&thread, NULL, client_func, (void*)passable_fd);
	}
}
